/* PostgreSQL.org Documentation Style */

/*
 * Documentation generated by XSL stylesheets has lower-case class
 * names, older documentation generated by DSSSL stylesheets has
 * upper-case class names, so we need to support both for a while.  In
 * some cases, the elements and classes differ further between the two
 * stylesheets.
 */

/* requires global.css, table.css and text.css to be loaded before this file! */
body {
  font-size: 76%;
}

.navheader table,
.NAVHEADER table {
  margin-left: 0;
}

/* Container Definitions */

#docContainerWrap {
  text-align: center; /* Win IE5 */
}

#docContainer {
  margin: 0 auto;
  width: 90%;
  padding-bottom: 2em;
  display: block;
  text-align: left; /* Win IE5 */
}

#docHeader {
  background-image: url("/media/img/docs/bg_hdr.png");
  height: 83px;
  margin: 0px;
  padding: 0px;
  display: block;
}

#docHeaderLogo {
  position: relative;
  width: 206px;
  height: 83px;
  border: 0px;
  padding: 0px;
  margin: 0 0 0 20px;
}

#docHeaderLogo img {
  border: 0px;
}

#docNavSearchContainer {
  padding-bottom: 2px;
}

#docNav, #docVersions {
  position: relative;
  text-align: left;
  margin-left: 10px;
  margin-top: 5px;
  color: #666;
  font-size: 0.95em;
}

#docSearch {
  position: relative;
  text-align: right;
  padding: 0;
  margin: 0;
  color: #666;
}

#docTextSize {
  text-align: right;
  white-space: nowrap;
  margin-top: 7px;
  font-size: 0.95em;
}

#docSearch form {
  position: relative;
  top: 5px;
  right: 0;
  margin: 0; /* need for IE 5.5 OSX */
  text-align: right; /* need for IE 5.5 OSX */
  white-space: nowrap; /* for Opera */
}

#docSearch form label {
  color: #666;
  font-size: 0.95em;
}

#docSearch form input {
  font-size: 0.95em;
}

#docSearch form #submit {
  font-size: 0.95em;
  background: #7A7A7A;
  color: #fff;
  border: 1px solid #7A7A7A;
  padding: 1px 4px;
}

#docSearch form #q {
  width: 170px;
  font-size: 0.95em;
  border:  1px solid #7A7A7A;
  background: #E1E1E1;
  color: #000000;
  padding: 2px;
}

.frmDocSearch {
  padding: 0;
  margin: 0;
  display: inline;
}

.inpDocSearch {
  padding: 0;
  margin: 0;
  color: #000;
}

#docContent {
  position: relative;
  margin-left: 10px;
  margin-right: 10px;
  margin-top: 40px;
}

#docFooter {
  position: relative;
  font-size: 0.9em;
  color: #666;
  line-height: 1.3em;
  margin-left: 10px;
  margin-right: 10px;
}

#docComments {
  margin-top: 10px;
}

#docClear {
  clear: both;
  margin: 0;
  padding: 0;
}

/* Heading Definitions */

h1, h2, h3 {
  font-weight: bold;
  margin-top: 2ex;
}

h1 {
  font-size: 1.4em;
}

h2 {
  font-size: 1.2em !important;
}

h3 {
  font-size: 1.1em;
}

h1 a:hover {
  color: #EC5800;
  text-decoration: none;
}

h2 a:hover,
h3 a:hover,
h4 a:hover {
  color: #666666;
  text-decoration: none;
}

/*
 * Change color of h2 chunk titles in XSL build.  (In DSSSL build,
 * these will be h1, which is already handled elsewhere.)
 */
.titlepage h2.title,
.refnamediv h2 {
  color: #EC5800;
}

/* Text Styles */

div.sect2,
div.SECT2 {
  margin-top: 4ex;
}

div.sect3,
div.SECT3 {
  margin-top: 3ex;
  margin-left: 3ex;
}

.txtCurrentLocation {
  font-weight: bold;
}

p, ol, ul, li {
  line-height: 1.5em;
}

.txtCommentsWrap {
  border: 2px solid #F5F5F5;
  width: 100%;
}

.txtCommentsContent {
  background: #F5F5F5;
  padding: 3px;
}

.txtCommentsPoster {
  float: left;
}

.txtCommentsDate {
  float: right;
}

.txtCommentsComment {
  padding: 3px;
}

#docContainer pre code,
#docContainer pre tt,
#docContainer pre pre,
#docContainer tt tt,
#docContainer tt code,
#docContainer tt pre {
  font-size: 1em;
}

pre.literallayout,
.screen,
.synopsis,
.programlisting,
.refsynopsisdiv p,
.caution,
.warning,
.note,
.tip,
.table table,
.informaltable table,
pre.LITERALLAYOUT,
.SCREEN,
.SYNOPSIS,
.PROGRAMLISTING,
.REFSYNOPSISDIV p,
table.CAUTION,
table.WARNING,
blockquote.NOTE,
blockquote.TIP,
table.CALSTABLE {
  -moz-box-shadow: 3px 3px 5px #DFDFDF;
  -webkit-box-shadow: 3px 3px 5px #DFDFDF;
  -khtml-box-shadow: 3px 3px 5px #DFDFDF;
  -o-box-shadow: 3px 3px 5px #DFDFDF;
  box-shadow: 3px 3px 5px #DFDFDF;
}

pre.literallayout,
.screen,
.synopsis,
.programlisting,
.refsynopsisdiv p,
.caution,
.warning,
.note,
.tip,
pre.LITERALLAYOUT,
.SCREEN,
.SYNOPSIS,
.PROGRAMLISTING,
.REFSYNOPSISDIV p,
table.CAUTION,
table.WARNING,
blockquote.NOTE,
blockquote.TIP {
  color: black;
  border-width: 1px;
  border-style: solid;
  padding: 2ex;
  margin: 2ex 0 2ex 2ex;
  overflow: auto;
  -moz-border-radius: 8px;
  -webkit-border-radius: 8px;
  -khtml-border-radius: 8px;
  border-radius: 8px;
}

pre.literallayout,
pre.synopsis,
pre.programlisting,
.refsynopsisdiv p,
.screen,
pre.LITERALLAYOUT,
pre.SYNOPSIS,
pre.PROGRAMLISTING,
.REFSYNOPSISDIV p,
.SCREEN {
  border-color: #CFCFCF;
  background-color: #F7F7F7;
}

.note,
.tip,
blockquote.NOTE,
blockquote.TIP {
  border-color: #DBDBCC;
  background-color: #EEEEDD;
  padding: 14px;
  width: 572px;
}

.note,
.tip,
.caution,
.warning,
blockquote.NOTE,
blockquote.TIP,
table.CAUTION,
table.WARNING {
  margin: 4ex auto;
}

.note p,
.tip p,
blockquote.NOTE p,
blockquote.TIP p {
  margin: 0;
}

.note pre,
.note code,
.tip pre,
.tip code,
blockquote.NOTE pre,
blockquote.NOTE code,
blockquote.TIP pre,
blockquote.TIP code {
  margin-left: 0;
  margin-right: 0;
  -moz-box-shadow: none;
  -webkit-box-shadow: none;
  -khtml-box-shadow: none;
  -o-box-shadow: none;
  box-shadow: none;
}

.caution,
.warning {
  max-width: 600px;
}

.tip h3,
.note h3,
.caution h3,
.warning h3 {
  text-align: center;
}

.emphasis,
.c2 {
  font-weight: bold;
}

.replaceable,
.REPLACEABLE {
  font-style: italic;
}

/* Table Styles */

table {
  margin-left: 2ex;
}

.table table td,
.table table th,
.informaltable table td,
.informaltable table th,
table.CALSTABLE td,
table.CALSTABLE th,
table.CAUTION td,
table.CAUTION th,
table.WARNING td,
table.WARNING th {
  border-style: solid;
}

.table table,
.informaltable table,
table.CALSTABLE,
table.CAUTION,
table.WARNING {
  border-spacing: 0;
  border-collapse: collapse;
}

.table table,
.informaltable table,
table.CALSTABLE
{
  margin: 2ex 0 2ex 2ex;
  background-color: #E0ECEF;
  border: 2px solid #A7C6DF;
}

.table table tr:hover td,
.informaltable table tr:hover td,
table.CALSTABLE tr:hover td
{
  background-color: #EFEFEF;
}

.table table td,
.informaltable table td,
table.CALSTABLE td {
  background-color: #FFF;
}

.table table td,
.table table th,
.informaltable table td,
.informaltable table th,
table.CALSTABLE td,
table.CALSTABLE th {
  border: 1px solid #A7C6DF;
  padding: 0.5ex 0.5ex;
}

table.CAUTION,
table.WARNING {
  border-collapse: separate;
  display: block;
  padding: 0;
  max-width: 600px;
}

.caution,
table.CAUTION {
  background-color: #F5F5DC;
  border-color: #DEDFA7;
}

.warning,
table.WARNING {
  background-color: #FFD7D7;
  border-color: #DF421E;
}

table.CAUTION td,
table.CAUTION th,
table.WARNING td,
table.WARNING th {
  border-width: 0;
  padding-left: 2ex;
  padding-right: 2ex;
}

table.CAUTION td,
table.CAUTION th {
  border-color: #F3E4D5
}

table.WARNING td,
table.WARNING th {
  border-color: #FFD7D7;
}

td.c1,
td.c2,
td.c3,
td.c4,
td.c5,
td.c6 {
  font-size: 1.1em;
  font-weight: bold;
  border-bottom: 0px solid #FFEFEF;
  padding: 1ex 2ex 0;
}

/* Link Styles */

#docNav a {
  font-weight: bold;
}

a:link,
a:visited,
a:active,
a:hover {
  text-decoration: underline;
}

a:link,
a:active {
  color:#0066A2;
}

a:visited {
  color:#004E66;
}

a:hover {
  color:#000000;
}

#docFooter a:link,
#docFooter a:visited,
#docFooter a:active {
  color:#666;
}

#docContainer code.function tt,
#docContainer code.FUNCTION tt {
  font-size: 1em;
}
